/*=============================================================================
Vitalnix/C-spark - User Management UI
user.c - Where the real work happens
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 1999 - 2004
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
  $Id: user.c,v 1.8 2004/03/11 17:55:11 hirogen2 Exp $
=============================================================================*/
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <libHX.h>
#include <accdb.h>
#include "cui/ds_entry.h"
#include "cui/global.h"
#include "cui/xml_in.h"

#define mm_sysctl(__req, __args...) \
     mm_output->sysctl((__req), mm_output->state, ## __args);

static int import_ds_Add(struct accdb_group *, FILE *);
static int import_ds_Delete(struct accdb_group *, FILE *);
static int read_input(void);
static int get_dgrp(struct accdb_group *);

static void cleanup_storage_addlist(void);
static void cleanup_storage_dellist(void);
static void free_addlist_entry(struct ds_entry *);
static void handle_user(struct accdb_user *, size_t *);
static char *mkhome(const char *, char *, size_t);
static FILE *open_logfile(void);

static struct HXbtree *storage_addlist = NULL;
static struct HXlld *storage_dellist = NULL;
static struct accdb_module *mm_output = NULL;

//-----------------------------------------------------------------------------
int db_open(void) {
    int eax;

    if((mm_output = accdb_load("shadow")) == NULL) {
        fprintf(stderr, "Could not load ACCDB back-end: %s\n",
         strerror(errno));
        return 0;
    }

    if((eax = mm_output->open(&mm_output->state)) <= 0) {
        fprintf(stderr, "Could not open ACCDB back-end: %s"
         " (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
        accdb_unload(mm_output);
        return 0;
    }

    return 1;
}

void db_close(void) {
    if(mm_output != NULL) {
        if(mm_output->state != NULL) { mm_output->close(mm_output->state); }
        accdb_unload(mm_output);
    }
    return;
}

void import_ds(void) {
    struct accdb_group dgroup;
    struct HXbtree *lnlist;
    FILE *logp;

    cleanup_storage_addlist();
    cleanup_storage_dellist();

    printf("\n"
      "IMPORT USERS\n"
      "============\n"
    );

    if(!get_dgrp(&dgroup)) { return; }
    if(read_input() <= 0) { return; }

    { /* And this is now the comparison algorithm, (half of) the core. While
      traversing the users, we mark used login names by putting them in an
      array (lnlist). Login names of users which are about to be deleted are
      also pushed there, so they do not get reassigned right now, but at the
      next run-through. */
      size_t users_grp = 0, users_keep = 0;
      struct accdb_user passwd;

      printf("\n" "Comparing Data Source against UDB\n");
      lnlist = HXbtree_init(HXBT_CPTR | HXBT_PCMP, strcmp);
      storage_dellist = HXlld_init(0);

      while(mm_output->usertrav(mm_output->state, &passwd)) {
          HXbtree_add(lnlist, passwd.lname);

          if((dgroup.gid == -1 || dgroup.gid == passwd.gid) &&
           (dgroup.gname == NULL || passwd.igrp == NULL ||
           strcmp(dgroup.gname, passwd.igrp) == 0)) {
              // Users we look for (only specified group)
              ++users_grp;
              handle_user(&passwd, &users_keep);
          }
      }

      printf("%u group member%s found\n", (unsigned int)users_grp,
       users_grp != 1 ? "s" : "");
      printf("   %u to keep\n", (unsigned int)users_keep);
      printf("   %lu to delete\n", storage_dellist->itemcount);
      printf("%lu to add\n", storage_addlist->itemcount);
    }

    if(storage_addlist->itemcount == 0 && storage_dellist->itemcount == 0) {
        printf("Already synchronized\n");
        goto import_ds__ret;
    }

    if(storage_addlist->itemcount > 0) {
        // Here comes the fixup loop (other part of the core).
        struct HXbtrav *travp = HXbtrav_init(storage_addlist, NULL);
        struct HXbtree_node *node;
        char buf[16];

        printf("Username fixup...\n");
        while((node = HXbtraverse(travp)) != NULL) {
            struct ds_entry *entry = node->data;
            int i = 0;
            buf[15] = '\0';
            strncpy(buf, entry->lname, 15);

            /* If the fixup'ed login name with the unique number already
            exists, bump the unique number. */
            while(HXbtree_find(lnlist, buf) != NULL) {
                fixup_login_name(entry->lname, ++i, buf, 16);
            }

            if(strcmp(entry->lname, buf) != 0) {
                HX_strclone(&entry->lname, buf);
            }
            HXbtree_add(lnlist, buf);
        }

        HXbtrav_free(travp);
    }

    HXbtree_free(lnlist);

    if(Optmap.c_aflush == AFLUSH_ON) {
        mm_sysctl(ACCDB_ADDFLAGS, ACCDB_AFLUSH);
    } else if(Optmap.c_aflush == AFLUSH_DEFAULT) {
        unsigned long f = mm_sysctl(ACCDB_ADDFLAGS, 0);
        printf("Notice: AUTOFLUSH is turned %s\n",
         (f & ACCDB_AFLUSH) ? "on" : "off");
    } else {
        mm_sysctl(ACCDB_DELFLAGS, ACCDB_AFLUSH);
    }

    if((logp = open_logfile()) == NULL) { goto import_ds__ret; }

    if(!import_ds_Add(&dgroup, logp) || !import_ds_Delete(&dgroup, logp)) {
        printf("Error during operation. You might need to redo the Import"
         " process\n" "to complete all jobs.\n");
    }

    /* We actually need to flush, since we have added home directores in /home,
    thus it would be nice to have at least some users have their account
    working. (Well, just keep a backup of your DB files, operations-so-far are
    recorded in the logfile and could thus be reversed easily.) */
    mm_sysctl(ACCDB_FLUSHDB);

import_ds__ret:
    cleanup_storage_addlist();
    cleanup_storage_dellist();
    return;
}

static int import_ds_Add(struct accdb_group *dgroup, FILE *logp) {
    // Adding the users to the system database
    ssize_t ic = storage_addlist->itemcount;
    struct HXbtree_node *node;
    struct HXbtrav *travp;
    char buf[MAXSNLEN], *plain_pw;
    int eax;

    if(ic == 0) { return 1; } // Nothing to do

    printf("\n"
      "Continue with adding users? (Yes/No)\n"
      "[Yes] > "
    );
    fgets(buf, MAXSNLEN, stdin);
    HX_chomp(buf);
    HX_strlower(buf);
    if(*buf != '\0' && *buf != 'y') {
        printf("Skipping addition\n");
        return 1;
    }

    if(Optmap.master_preadd != NULL) {
        char cmd[MAXLNLEN];
        snprintf(cmd, MAXLNLEN, Optmap.master_preadd, (unsigned int)ic);
        system(cmd);
    }

    plain_pw = malloc(Optmap.i_pswdlen + 1);
    travp = HXbtrav_init(storage_addlist, NULL);
    while((node = HXbtraverse(travp)) != NULL) {
        struct ds_entry *in = node->data;
        struct accdb_user out;
        char uhome[MAXFNLEN];

        memset(&out, 0, sizeof(struct accdb_user));
        out.lname = in->lname;
        out.uid   = -1;
        out.gid   = dgroup->gid;
        out.gecos = node->key;
        out.home  = mkhome(in->lname, uhome, MAXFNLEN);
        out.shell = Optmap.f_shell;
        out.igrp  = dgroup->gname;
        out.sgrp  = NULL;

        if(Optmap.i_pswdlen < 0) {
            out.passwd = "!";
            out.last_change = 0;
        } else if(Optmap.i_pswdlen == 0) {
            out.passwd = "";
            out.last_change = (unsigned long)time(NULL) / 86400; // "now"
        } else {
            unsigned long fl = GENPW_ONE_DIGIT | GENPW_ONE_CASE;
            if(Optmap.b_pwphon) { fl |= GENPW_PHONEMIC; }
            accdb_genpw(plain_pw, Optmap.i_pswdlen, fl);
            accdb_cryptpw(plain_pw, NULL, Optmap.c_crymeth, &out.passwd);
            out.last_change = (unsigned long)time(NULL) / 86400;
        }

        out.keep_min    = 0;
        out.keep_max    = 366;
        out.warn_age    = 21;
        out.expire      = -1;
        out.inactive    = -1;

        if(Optmap.user_preadd != NULL) {
            /* Order is important, as we need to stay compatible in
            future. See SOURCEROOT/vetc/spark.cfg for the assigned
            order. */
            char cmd[MAXLNLEN];
            snprintf(cmd, MAXLNLEN, Optmap.user_preadd,
             out.lname, out.uid, out.gid, out.igrp, out.sgrp,
             out.gecos, out.home, out.shell);
            system(cmd);
        }

        if((eax = mm_output->useradd(mm_output->state, &out)) <= 0) {
            memset(plain_pw, '\0', Optmap.i_pswdlen + 1);
            free(plain_pw);
            printf("\n" "useradd: Something went wrong. (eax=%d,"
             " errno=%d: %s)\n", eax, errno, strerror(errno));
            HXbtrav_free(travp);
            return 0;
        }

        if(logp != NULL) {
            fprintf(logp, "+:%s:%s:%s:%s\n", in->sgroup, in->rname,
             out.lname, plain_pw);
        }

        if(Optmap.i_pswdlen > 0) { free(out.passwd); }

        // Create home directory and optionally copy skeleton dir
        if(!create_home(out.home, out.uid, out.gid)) {
            fprintf(stderr, "\e[2K" "Warning: Could not create home directory"
             " %s\n", strerror(errno));
        }

        if(Optmap.user_postadd != NULL) {
            char cmd[MAXLNLEN];
            snprintf(cmd, MAXLNLEN, Optmap.user_postadd, out.lname);
            if(Optmap.c_aflush == AFLUSH_POSTADD) { mm_sysctl(ACCDB_FLUSHDB); }
            system(cmd);
        }

        fprintf(stderr, "\e[2K" "add: %u users remaining\r",
         (unsigned int)--ic);
    }

    memset(plain_pw, '\0', Optmap.i_pswdlen + 1);
    free(plain_pw);
    HXbtrav_free(travp);

    if(Optmap.master_postadd != NULL) {
        char cmd[MAXLNLEN];
        snprintf(cmd, MAXLNLEN, Optmap.master_postadd,
         (unsigned int)storage_addlist->itemcount);
    }

    fprintf(stderr, "\e[2K" "Successfully added %u users\n",
     (unsigned int)storage_addlist->itemcount);
    return 1;
}

static int import_ds_Delete(struct accdb_group *dgroup, FILE *logp) {
    // Deleting the old users
    size_t ic = storage_dellist->itemcount;
    struct HXlld_node *travp;
    char buf[MAXSNLEN];
    int eax;

    if(ic == 0) { return 1; } // Nothing to do

    printf("\n"
      "Continue with deleting users? (Yes/No)\n"
      "[Yes] > "
    );
    fgets(buf, MAXSNLEN, stdin);
    HX_chomp(buf);
    HX_strlower(buf);
    if(*buf != '\0' && *buf != 'y') {
        printf("Skipping deletion\n");
        return 1;
    }

    /* We generate a time stamp here so that a possible backup operation will
    write into the same directory for all users to be deleted. The format
    should also stay the same (compatibility reasons), if possible. */
    {
      time_t now = time(NULL);
      struct tm res;
      strftime(buf, MAXSNLEN, "%Y%m%d-%H%M%S", localtime_r(&now, &res));
    }

    if(Optmap.master_predel != NULL) {
        char cmd[MAXLNLEN];
        snprintf(cmd, MAXLNLEN, Optmap.master_predel,
         (unsigned int)ic, buf);
        system(cmd);
    }

    travp = storage_dellist->first;
    while(travp != NULL) {
        struct accdb_user rq;
        char *lname;

        memset(&rq, 0, sizeof(struct accdb_user));
        rq.lname = travp->ptr;
        rq.uid = rq.gid = -1;

        if((eax = mm_output->userinfo(mm_output->state, &rq, &rq, 1)) < 0) {
            printf("\n" "userinfo: Something went wrong. (eax=%d, "
             "errno=%d: %s)\n", eax, errno, strerror(errno));
            return 0;
        } else if(eax == 0) {
            // This should not happen. Someone deleted the user before we did.
            printf("\n" "Warning: Someone deleted \"%s\" before we did.\n",
             rq.lname);
            continue;
        }

        if(Optmap.user_predel != NULL) {
            char cmd[MAXLNLEN];
            snprintf(cmd, MAXLNLEN, Optmap.user_predel, rq.lname,
             rq.uid, rq.gid, rq.igrp, rq.home, buf);
            system(cmd);
        }
        HX_rrmdir(rq.home);

        // need a copy since travp->ptr will be undefined after userdel
        lname = HX_strdup(travp->ptr);

        // (surprise, we can reuse "rq" without modification)
        if((eax = mm_output->userdel(mm_output->state, &rq)) < 0) {
            printf("\n" "userdel: Something went wrong. (eax=%d, "
             "errno=%d: %s)\n", eax, errno, strerror(errno));
            return 0;
        }

        if(logp != NULL) { fprintf(logp, "-:%s\n", lname); }

        if(Optmap.user_postdel != NULL) {
            char cmd[MAXLNLEN];
            snprintf(cmd, MAXLNLEN, Optmap.user_postdel, lname);
            system(cmd);
        }

        free(lname);
        fprintf(stderr, "\e[2K" "delete: %u users remaining\r",
         (unsigned int)--ic);
        travp = travp->Next;
    }

    if(Optmap.master_postdel != NULL) {
        char cmd[MAXLNLEN];
        snprintf(buf, MAXLNLEN, Optmap.master_postdel,
         (unsigned int)storage_dellist->itemcount, buf);
        system(cmd);
   }

    fprintf(stderr, "\e[2K" "Successfully deleted %u users\n",
     (unsigned int)storage_dellist->itemcount);
    return 1;
}

static int get_dgrp(struct accdb_group *grp) {
    /* Here, we request a group to which compare the data source against.
    Usually, a group is always the same for a particular input set otherwise
    strange things may happen. You can either enter the group name or a GID.
    (Once there is a non-digit in your input, it is treated as a group
    name.) */
    char buf[MAXSNLEN];
    printf("\n"
      "Enter the group to which compare users from the Data Source with.\n"
      "Destianation group > "
    );

    fgets(buf, MAXSNLEN, stdin);
    HX_chomp(buf);
    if(*buf == '\0') { return 0; }

    if(only_digits(buf)) {
        grp->gid = strtol(buf, NULL, 0);
        grp->gname = NULL;
    } else {
        grp->gid = -1;
        grp->gname = buf;
    }

    if(!mm_output->groupinfo(mm_output->state, grp, grp, 1)) {
        printf("Group \"%s\" not found\n", buf);
        return 0;
    }

    return 1;
}

static int read_input(void) {
    /* In this part, we stuff all users from the data source into an
    associative array, called storage_addlist. Some elements of that array will
    further be changed until it is ready for running through import_ds(). */
    struct ds_entry *bucket;
    char buf[MAXLNLEN], fbuf[MAXFNLEN];
    void *state;
    int eax;

    printf("\n"
      "Please enter the pathname to the file serving as input. The file must\n"
      "be in XML and must have the tags described in doc/sp_xmlds.html.\n"
      "Input file > "
    );
    fgets(fbuf, MAXFNLEN, stdin);
    HX_chomp(fbuf);

    if(*fbuf == '\0') { return 0; }
    if((eax = xmlds_open(fbuf, &state)) <= 0) { return eax; }

    /* This loop fills the storage_addlist assoc array by generating a suited
    (pre-stage) login name.
    Pre-stage login names: these login names may already be used by the system
    and must first be fixed up (i.e. by adding a unique postcount number), so
    that there is no double login name in the user database. */

    storage_addlist = HXbtree_init(HXBT_ASSOC);

    while(1) {
        bucket = malloc(sizeof(struct ds_entry));

        if((eax = xmlds_read(state, bucket)) < 0) {
            fprintf(stderr, "Warning: Stumbled upon a bogus entry"
             " in Data Source, continuing\n");
            free(bucket);
            continue;
        } else if(eax == 0) {
            // End of list
            free(bucket);
            break;
        }

        bucket->lname = make_login_name(bucket->nname, bucket->vname);

        if(bucket->nname != NULL) {
            snprintf(buf, MAXLNLEN, "%s %s", bucket->vname, bucket->nname);
            bucket->rname = HX_strdup(buf);
        } else {
            bucket->rname = HX_strdup(bucket->vname);
        }

        if(bucket->xuid != NULL) {
            snprintf(buf, MAXLNLEN, "%s,%s", bucket->rname, bucket->xuid);
        } else {
            buf[MAXLNLEN - 1] = '\0';
            strncpy(buf, bucket->rname, MAXLNLEN - 1);
        }
        HXbtree_add(storage_addlist, buf, bucket);
    }

    xmlds_close(state);
    return 1;
}

//-----------------------------------------------------------------------------
static void cleanup_storage_addlist(void) {
    if(storage_addlist != NULL) {
        struct HXbtrav *travp = HXbtrav_init(storage_addlist, NULL);
        struct HXbtree_node *node;

        while((node = HXbtraverse(travp)) != NULL) {
            free_addlist_entry(node->data);
        }

        HXbtrav_free(travp);
        HXbtree_free(storage_addlist);
        storage_addlist = NULL;
    }
    return;
}

static void cleanup_storage_dellist(void) {
    if(storage_dellist != NULL) {
        HXlld_free(storage_dellist);
        storage_dellist = NULL;
    }
    return;
}

static void free_addlist_entry(struct ds_entry *e) {
#define FREE(__p) if((__p) != NULL) { free(__p); }
    if(e == NULL) { return; }
    FREE(e->vname);
    FREE(e->nname);
    FREE(e->xuid);
    FREE(e->sgroup);
    FREE(e->lname);
    FREE(e->rname);
    free(e);
    return;
#undef FREE
}

static void handle_user(struct accdb_user *p, size_t *ukeep) {
    char buf[MAXLNLEN], *brkp;

    buf[MAXLNLEN - 1] = '\0';
    strncpy(buf, p->gecos, MAXLNLEN - 1);

    if((brkp = strchr(buf, ',')) != NULL &&
     (brkp = strchr(++brkp, ',')) != NULL) {
        // strip further fields
        *brkp = '\0';
    }

    if(HXbtree_find(storage_addlist, buf) != NULL) {
        ++*ukeep;
        free_addlist_entry(HXbtree_get(storage_addlist, buf));
        HXbtree_del(storage_addlist, buf);
        // what remains in storage_addlist is the users to add
    } else {
        // not anymore in data source
        HXlld_push(storage_dellist, p->lname);
    }

    return;
}

static char *mkhome(const char *ln, char *buf, size_t s) {
    if(Optmap.c_split == 1 || strlen(ln) == 1) {
        snprintf(buf, s, "%s/%c/%s", Optmap.f_home, *ln, ln);
    } else if(Optmap.c_split >= 2) {
        snprintf(buf, s, "%s/%c/%c%c/%s",
         Optmap.f_home, *ln, ln[0], ln[1], ln);
    } else {
        snprintf(buf, s, "%s/%s", Optmap.f_home, ln);
    }
    return buf;
}

static FILE *open_logfile(void) {
    char logfile[MAXFNLEN];
    FILE *rv;

    printf("\n"
      "Enter a name for a log file containing a preliminary list of\n"
      "added users and their respective plaintext passwords.\n"
      "Log file > "
    );

    fgets(logfile, MAXFNLEN, stdin);
    HX_chomp(logfile);
    if(*logfile == '\0') {
        printf("\n" "No filename given, aborting.\n");
        return NULL;
    }

    if(file_exists(logfile)) {
        char buf[MAXLNLEN];
        printf("\n"
          "File already exists. Specify what to do: [O]verwrite, "
          "[A]ppend or [C]ancel?\n"
          "[Cancel] > "
        );
        fgets(buf, MAXLNLEN, stdin);
        HX_strlower(buf);
        HX_chomp(buf);
        if(*buf == 'a') { rv = fopen(logfile, "a"); }
        else if(*buf == 'o') { rv = fopen(logfile, "w"); }
        else { return NULL; }
    } else {
        rv = fopen(logfile, "w");
    }

    if(rv == NULL) {
        fprintf(stderr, "Error: Could not open %s: %s\n",
         logfile, strerror(errno));
    }

    return rv;
}

//==[ End of file ]============================================================
