/*=============================================================================
hxtools:gxmalloc
  Copyright (C) Jan "Hirogen2" Engelhardt <jengelh [at] linux01 gwdg de>,
  2003 - 2004
  -- License restrictions apply (Dual FAL/LGPL2).
  -- For details see doc/FAL.txt and doc/LGPL2.txt.
  http://linux01.org:2222/f/hxtools.tbz2
===============================================================================

DESCRIPTION
    GXmalloc is a memory leak detection helper utility, or as you would
    probably say, a memory debugger. It is intended to be as free as possible.
    You do not need to change any of your code to get it working. Just pass
    -lgxmalloc to the linker, and that is it.
    (You may need additional options like -L and -Wl,-rpath, to point gcc-ld to
    GXmalloc's path if it is not in the default system directories.)

FEATURES
    - find unfreed objects
    - find "attempts to free an object twice" (and "attempts to free NULL")
    - pre-/post-fence write detection (between 8 and 32 bytes at front and
      back, depending on architecture); it also uses a dynamic signature so if
      a fence breach is not detected in one run, it is with the next.
    - good stack backtrace in objects compiled without -g
      (you will still need a symbol table, so do not strip)
    - perfect stack backtrace in objects compiled with -g or more
    - totally thread-safe (or so I believe)

    At the moment, it does not do anymore, as these are the main leaks that
    hide in applications. I did not need any other functionality yet, but if
    you do, let me know - maybe it is something useful for me too.

RESTRICTIONS
    Because of how GXmalloc is laid out, it only works with GNU libc, GNU cc
    and GNU ld, but in turn effectively uses GNU's features. Whether it works
    on any architecture other than x86 and x86_64 (AMD64) if you took out the
    x86 specific parts, is unknown.

COMPILING GXMALLOC
    The Makefile usually does it for you, and the only note I am adding here
    (even for myself) is that GXmalloc cannot be compiled -fomit-frame-pointers
    or it will segfault (in get_ebp()). (I know why, thanks :-)

    You will need libHX, which serves as the tactical backbone for GXmalloc,
    including RB/B-tree and some string functions.
    http://vitalnix.sf.net/libHX/

    If you need a deeper stack backtrace than the resulting string buffer is
    long, increase SBT_SIZE in backtrace().

OPTIONS
    Options are set via the environment variable GXMALLOC; they are separated
    by comma.


    active    Activate GXmalloc (so you can always have -lgxmalloc linked in)
              There is no overhead when GXmalloc is not activated.

    fence     Activate pre-fence and post-fence write detection

    flush     Activate autoflush for stdout, stderr. (Note that GXmalloc is ran
              before any of your code, so if you setvbuf() later, you will
              override this.)

    line      Tries to resolve addresses to file lines. In a stack backtrace
              you might see
                  read_options (opt.c:567) < main (init.c:3) < _init (0x...)
              If the exact line can not be resolved, falls back to the
              instruction address (as with _init in this example).

    noansi    Do not use ANSI color to markup GXM messages

    verbose   Print general messages

    vmf       Print malloc()/calloc(), realloc() and free() calls together with
              the total allocated object count.


    $ gcc -o prog prog.c -lgxmalloc -L. -Wl,-rpath,.
    $ export GXMALLOC=active,fence,noansi,verbose
    $ ./prog

CHANGELOG
    March 14 2003
    + added to the documentation and comments
    + itemcount was %ld in printf() statements, should be %lu
    + realloc_hook() allocated too much if we ran without fence checking
    + removed the unused "pr" parameter of check_fence()
    + load_sfile(): added quotes around %s
    + added vmf option
    + relaxed entry locking for _GXM_init
    + "perfect" stack backtrace (line resolving)
    + "unfreed object" messages displayed a wrong address if fence tracking
      was enabled
    + added stats (max alloc obj count, malloc/etc. ops)

    March 08 2003
    + added a "maximum allocated objects" counter
    + "s" was missing as arg to printf in _GXM_free_hook

    December 25 2003
    + fence is now aligned at dword boundary (+ fence checking within pad area)
    + replaced the fence checking code with something even more optimized by
      exploting architecture-independent hardware facts.
    + new option parser
-----------------------------------------------------------------------------*/
#define _GNU_SOURCE 1
#include <sys/types.h>
#include <ctype.h>
#include <errno.h>
#include <malloc.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libHX.h>

typedef unsigned char byte;
/* dammit, "ulong" was already taken, but is not guaranteed to be portable.
Then again, GXmalloc is bound to GNU, so... */
typedef unsigned long Ulong;

#define __ROUND_UP(__x, __n) (((__x) + (__n) - 1) & ~((__n) - 1))
#define MAXFNLEN 512
#define MAXLNLEN 1024
#define _weak_alias(__name, __aname) \
  extern __typeof(__name) __aname __attribute__((weak, alias(#__name)));
#define _strong_alias(__name, __aname) \
  extern __typeof(__name) __aname __attribute__((alias(#__name)));

enum {
    FENCE_PRE  = 1 << 0,
    FENCE_POST = 1 << 1,
};

// Function prototypes
extern void *__libc_malloc(size_t);
extern void *__libc_calloc(size_t, size_t);
extern void *__libc_realloc(void *, size_t);
extern void __libc_free(void *);

extern void _GXM_init(const void *) __attribute__((constructor));
extern void _GXM_deinit(void) __attribute__((destructor));
extern void *_GXM_malloc(size_t);
extern void *_GXM_calloc(size_t, size_t);
extern void *_GXM_realloc(void *, size_t);
extern void _GXM_free(void *);
extern void *_GXM_malloc_hook(size_t, const void *);
extern void *_GXM_realloc_hook(void *, size_t, const void *);
extern void _GXM_free_hook(void *, const void *);

static char *backtrace(void);
static int check_dlopen(void);
static int check_fence(const int *);
static void check_fence_all(void);
static void dump_unfreed(void);
static const void *get_ebp(int);
static void get_flags(const char *);
static void hooks_deploy(void);
static void hooks_restore(void);
static int load_sfile(const char *, Ulong);
static int gprint(int, char *, ...);
static const char *rsymbol(const void *);
static char *rsymbol_f(const void *);
static void update_maps(void);
static int value_cmp(void *, void *);

// Variables
static struct {
    byte active, fence, flush, noansi, verbose, vmf, line;
    char *logf;
    FILE *logp;
} Flags;

static struct {
    void *(*omalloc)(size_t, const void *);
    void *(*orealloc)(void *, size_t, const void *);
    void (*ofree)(void *, const void *);
    struct HXbtree *aobj, *fmap, *amap, *stab;
    pthread_mutex_t lock;
    int inited, sid;
} GV = {
    lock: PTHREAD_MUTEX_INITIALIZER, inited: 0, sid: 0xDEADCAC0,
};

static struct {
    Ulong op_malloc, op_realloc, op_free, maxobj;
} Stats;

//-----------------------------------------------------------------------------
void _GXM_init(const void *caller) {
    void *rd;

    if(GV.inited) { return; }
    pthread_mutex_lock(&GV.lock);
    if(GV.inited++) { return; }

    memset(&Stats, 0, sizeof(Stats));
    get_flags(getenv("GXMALLOC"));
    if(!Flags.active) {
        GV.inited = 0;
        return;
    }

    rd = HX_rand_init(HXRAND_DEV_URANDOM | HXRAND_SYS_RAND);
    if(rd != NULL) {
        byte *sp = (byte *)&GV.sid, i = sizeof(int);
        while(i--) { *sp++ = HX_rand(rd); }
        HX_rand_deinit(rd);
    } else {
        memcpy(&GV.sid, "GXMALLOC2004", sizeof(int));
    }

    GV.aobj = HXbtree_init(HXBT_ASSOC | HXBT_LKEY | HXBT_PCMP, value_cmp);
    GV.fmap = HXbtree_init(HXBT_CPTR | HXBT_PCMP, strcmp);
    GV.amap = HXbtree_init(HXBT_ASSOC | HXBT_LKEY | HXBT_CPTR | HXBT_PCMP,
     value_cmp);
    GV.stab = HXbtree_init(HXBT_ASSOC | HXBT_LKEY | HXBT_CPTR | HXBT_PCMP,
     value_cmp);

    GV.omalloc = __malloc_hook;
    GV.orealloc = __realloc_hook;
    GV.ofree = __free_hook;

    if(Flags.flush) {
        setvbuf(stdout, NULL, _IONBF, 0);
        setvbuf(stderr, NULL, _IONBF, 0);
    }

    update_maps();
    gprint(1, "GXmalloc ready\n");

    atexit(_GXM_deinit);
    pthread_mutex_unlock(&GV.lock);
    hooks_deploy();
    return;
}

void _GXM_deinit(void) {
    if(GV.inited <= 0) { return; }
    pthread_mutex_lock(&GV.lock);
    if(GV.inited-- <= 0) { return; }

    if(!Flags.active) { return; }
    hooks_restore();
    gprint(1, "GXmalloc hooks deinitialized\n");

    gprint(Flags.verbose, "Statistics:\n");
    gprint(Flags.verbose, "    Max alloc obj count: %7lu\n", Stats.maxobj);
    gprint(Flags.verbose, "    malloc() calls:      %7lu\n", Stats.op_malloc);
    gprint(Flags.verbose, "    realloc() calls:     %7lu\n", Stats.op_realloc);
    gprint(Flags.verbose, "    free() calls:        %7lu\n", Stats.op_free);
    gprint(Flags.verbose, "    Total transactions:  %7lu\n", Stats.op_malloc + Stats.op_realloc + Stats.op_free);
    gprint(Flags.verbose, "    Unfreed pointers:    %7lu\n", GV.aobj->itemcount);

    if(GV.aobj->itemcount > 0) {
        check_fence_all();
        dump_unfreed();
    }

    HXbtree_free(GV.aobj);
    HXbtree_free(GV.fmap);
    HXbtree_free(GV.amap);
    HXbtree_free(GV.stab);

    pthread_mutex_unlock(&GV.lock);
    pthread_mutex_destroy(&GV.lock);
    return;
}

//-----------------------------------------------------------------------------
/* Re-wrappers

A weird net of functions, which all lead to Glibc's malloc.

Other memory leak debuggers have their own malloc() -- but doing so does not
make libc functions use the new malloc(), because libc's internal functions
all call __libc_malloc(). So the way to go is to hook into __libc_malloc()
itself, which provides __malloc_hook(3) for our pleasure! */

void *_GXM_malloc(size_t s) {
    return __libc_malloc(s);
}

void *_GXM_calloc(size_t n, size_t s) {
    return __libc_calloc(n, s);
}

void *_GXM_realloc(void *ptr, size_t s) {
    return __libc_realloc(ptr, s);
}

void _GXM_free(void *ptr) {
    __libc_free(ptr);
    return;
}

_strong_alias(_GXM_malloc, malloc);
_strong_alias(_GXM_calloc, calloc);
_strong_alias(_GXM_realloc, realloc);
_strong_alias(_GXM_free, free);

//-----------------------------------------------------------------------------
// Libc hooks

void *_GXM_malloc_hook(size_t s, const void *caller) {
    int *rp, *op = NULL; // real pointer, out pointer

    pthread_mutex_lock(&GV.lock);
    hooks_restore();
    if(check_dlopen()) { update_maps(); }

    if(Flags.fence) {
        /* If we want to use fence checking, we need an extra header so that
        check_fence knows how long "data" is. */

        rp = __libc_malloc(__ROUND_UP(s + 6 * sizeof(int), sizeof(int)));
        if(rp != NULL) {
            /* [SID] [BLOCKLEN] [SID] [SID] [...data...] [pad] [SID] [SID]
            Detectable:  4*sizeof(int)             at the front and
                         2*sizeof(int)+sizeof(pad) at the end
            */
            int *ff   = (int *)__ROUND_UP((Ulong)(rp + 4) + s,
                        sizeof(int));
            *(rp + 1) = s;
            *rp       =
            *(rp + 2) =
            *(rp + 3) =
            /* if s==0 then ff-1==rp+3, otherwise it is the pad block
            We can safely do ff-1 (possibly partially data area!), as
            there is no data written to it yet. */
            *(ff - 1) =
            *ff       =
            *(ff + 1) = GV.sid;

            op = rp + 4;
            HXbtree_add(GV.aobj, rp, backtrace());
        }
    } else if((op = __libc_malloc(s)) != NULL) {
        HXbtree_add(GV.aobj, op, backtrace());
    }

    gprint(Flags.vmf, "[objs=%5lu] 0x%08lX = malloc(s=%lu)\n",
     GV.aobj->itemcount, (Ulong)op, (Ulong)s);

    ++Stats.op_malloc;
    if(GV.aobj->itemcount > Stats.maxobj) {
        Stats.maxobj = GV.aobj->itemcount;
    }

    hooks_deploy();
    pthread_mutex_unlock(&GV.lock);
    return op;
}

void *_GXM_realloc_hook(void *ip, size_t s, const void *caller) {
    int *rp = NULL, *np, *op = NULL;

    pthread_mutex_lock(&GV.lock);
    hooks_restore();

    if(ip != NULL) {
        rp = ip - 4;
        if(Flags.fence) { check_fence(rp); }
    }

    if(Flags.fence && s != 0) {
        np = __libc_realloc(rp, __ROUND_UP(s + 6 * sizeof(int), sizeof(int)));
    } else {
        np = __libc_realloc(rp, s);
    }

    if(np != rp) {
        if(rp != NULL) { HXbtree_del(GV.aobj, rp); }
        if(np != NULL) { HXbtree_add(GV.aobj, np, backtrace()); }
        rp = np;
    }

    if(Flags.fence && rp != NULL) {
        /* Need to adjust the trailing two SID blocks.
        Eh, let 's update all blocks ... for stability-security. */
        void *ps = (void *)(rp + 4) + s;
        int *ff  = (int *)__ROUND_UP((Ulong)ps, sizeof(int));
        int l    = (void *)ff - ps;

        *(rp + 1) = s;
        *rp       =
        *(rp + 2) =
        *(rp + 3) =
        *ff       =
        *(ff + 1) = GV.sid;

        /* Need to take care not to overwrite data! The pad area might be
        smaller than sizeof(int). */
        memcpy(ps, (void *)&GV.sid + sizeof(int) - l, l); // pad filling
        op = rp + 4;
    }

    gprint(Flags.vmf, "[objs=%5lu] 0x%08lX = realloc(ptr=0x%08lX,"
     " s=%lu)\n", GV.aobj->itemcount, (Ulong)op, (Ulong)ip, (Ulong)s);

    ++Stats.op_realloc;
    if(GV.aobj->itemcount > Stats.maxobj) {
        Stats.maxobj = GV.aobj->itemcount;
    }

    hooks_deploy();
    pthread_mutex_unlock(&GV.lock);
    return op;
}

void _GXM_free_hook(void *ip, const void *caller) {
    int *rp = NULL; // in pointer, real pointer

    pthread_mutex_lock(&GV.lock);
    hooks_restore();
    if(ip != NULL) { rp = ip - 4; }
    ++Stats.op_free;

    if(rp == NULL) {
        char *s = backtrace();
        gprint(1, "Tried to free NULL {Trace: %s}\n", s);
        free(s);
        goto __end_free_hook;
    }

    if(HXbtree_del(GV.aobj, rp) == NULL && errno == ENOENT) {
        char *s = backtrace();
        gprint(1, "Tried to free 0x%08lX twice {Trace: %s}\n", (Ulong)ip, s);
        free(s);
        goto __end_free_hook;
    }

    if(Flags.fence) { check_fence(rp); }
    __libc_free(rp);
    gprint(Flags.vmf, "[objs=%5lu] free(ptr=0x%08lX)\n",
     GV.aobj->itemcount, (Ulong)ip);

__end_free_hook:
    hooks_deploy();
    pthread_mutex_unlock(&GV.lock);
    return;
}

//-----------------------------------------------------------------------------
// Internal functions

static char *backtrace(void) {
#define SBT_SIZE 256 /* maximum buffer size for backtrace dumps */
    /* Skip 4 stack frames: backtrace, _GXM_malloc_hook, __libc_malloc and
    _GXM_malloc. */
    const void *ebp = get_ebp(4);
    char buf[SBT_SIZE], ibuf[SBT_SIZE];
    int lv = 0;
    memset(buf, '\0', SBT_SIZE);

    while(ebp != NULL) {
        const void *f_addr = *(const void **)(ebp + sizeof(void *));
        const char *sym = rsymbol(f_addr);

        if(sym == NULL) {
            snprintf(ibuf, SBT_SIZE, "0x%08lX", (Ulong)f_addr);
        } else if(lv++ < 2) {
            char *line;
            if(Flags.line && (line = rsymbol_f(f_addr)) != NULL) {
                snprintf(ibuf, SBT_SIZE, "%s(%s)", sym, line);
                free(line);
            } else {
                snprintf(ibuf, SBT_SIZE, "%s(0x%08lX)", sym, (Ulong)f_addr);
            }
        } else {
            snprintf(ibuf, SBT_SIZE, "%s", sym);
        }

        HX_strlcat(buf, ibuf, SBT_SIZE);
        ebp = *(const void **)ebp;
        if(ebp != NULL) { HX_strlcat(buf, " < ", SBT_SIZE); }
    }

    return HX_strdup(buf);
#undef SBT_SIZE
}

static int check_dlopen(void) {
    /* Checks whether this malloc() is due to a dlopen() call. We exploit the
    fact that dlopen() acquires some space after having loaded a library, and
    thus can load the symbols of that lib. */
    const void *ebp = get_ebp(4);
    int found = 0;

    while(ebp != 0x0) {
        const char *sym = rsymbol(*(const void **)(ebp + sizeof(void *)));
        if(sym != NULL && strncmp(sym, "dlopen@@", 8) == 0) { ++found; break; }
        ebp = *(const void **)ebp;
    }

    return found;
}

static int check_fence(const int *rp) {
    // Checks whether SID blocks of an (RP) object are intact
    int *ff, ret = 0;
    size_t s = *(rp + 1);
    void *ps = (void *)(rp + 4) + s;
    int l;

    if(*rp != GV.sid || *(rp + 2) != GV.sid || *(rp + 3) != GV.sid) {
        ret |= FENCE_PRE;
        /* Using gprint(1) since check_fence() is anyway only called if
        Flags.fence is on. */
        gprint(1, "    Pre-fence broken for object 0x%08lX\n", (Ulong)(rp + 4));
    }

    ff = (int *)__ROUND_UP((Ulong)ps, sizeof(int));
    l = (void *)ff - ps;

    if(memcmp(ps, (void *)&GV.sid + sizeof(int) - l, l) != 0 ||
     *ff != GV.sid || *(ff + 1) != GV.sid) {
        ret |= FENCE_POST;
        gprint(1, "    Post-fence broken for object 0x%08lX\n", (Ulong)(rp + 4));
    }

    return ret;
}

static void check_fence_all(void) {
    /* Walks through the tree of allocated objects and checks each for fence
    overrun. */
    struct HXbtrav *travp = HXbtrav_init(GV.aobj, NULL);
    struct HXbtree_node *n;

    gprint(1, "Fence Check of all objects:\n");
    while((n = HXbtraverse(travp)) != NULL) { check_fence(n->key); }
    HXbtrav_free(travp);
    return;
}

static void dump_unfreed(void) {
    /* Prints out all objects which are allocated at the time dump_unfreed()
    runs through. */
    struct HXbtrav *travp = HXbtrav_init(GV.aobj, NULL);
    struct HXbtree_node *n;

    gprint(1, "Table of unfreed objects:\n");
    while((n = HXbtraverse(travp)) != NULL) {
        const int *addr = n->key;
        if(Flags.fence) { addr += 4; }
        gprint(1, "    0x%08lX {Trace: %s}\n",
         (Ulong)addr, n->data);
    }

    HXbtrav_free(travp);
    return;
}

static const void *get_ebp(int skip) {
    const void *ebp = NULL;
#if defined(__x86_64__)
    asm volatile("movq %%rbp, %0;" :"=m"(ebp));
#else
    // EBP works on X86_64, too.
    asm volatile("movl %%ebp, %0;" :"=m"(ebp));
#endif

    while(skip--) {
        if(ebp == NULL) { return NULL; }
        ebp = *(const void **)ebp;
    }

    return ebp;
}

static void get_flags(const char *i) {
#define GET_FLAG(__s) if(strcmp(cp, (#__s)) == 0) { Flags.__s = 1; }
    char buf[MAXLNLEN], *bufp = buf, *cp;
    memset(&Flags, 0, sizeof(Flags));
    if(i == NULL) { return; }

    buf[MAXLNLEN - 1] = '\0';
    strncpy(buf, i, MAXLNLEN - 1);

    while((cp = HX_strsep(&bufp, ",")) != NULL) {
        char *v = strchr(cp, '=');
        if(v != NULL) { *v++ = '\0'; }
        GET_FLAG(active);
        GET_FLAG(fence);
        GET_FLAG(flush);
        GET_FLAG(noansi);
        GET_FLAG(verbose);
        GET_FLAG(vmf);
        GET_FLAG(line);
    }

    return;
#undef GET_FLAG
}

static int gprint(int f, char *fmt, ...) {
    int rv = 0;
    va_list argp;
    va_start(argp, fmt);

    if(f) {
        if(!Flags.noansi && isatty(fileno(stderr))) {
            fprintf(stderr, "\e[31mGXM\e[1m>\e[0m "); // ]]]
        } else {
            fprintf(stderr, "GXM> ");
        }
        rv = vfprintf(stderr, fmt, argp);
    }

    va_end(argp);
    return rv;
}

/* The following two functions make up 30% of GXmalloc's speed, because we
use (G)libc's malloc -- which is one of the fastest, since... it's Glibc. */

static void hooks_deploy(void) {
    __malloc_hook = _GXM_malloc_hook;
    __realloc_hook = _GXM_realloc_hook;
    __free_hook = _GXM_free_hook;
    return;
}

static void hooks_restore(void) {
    __malloc_hook = GV.omalloc;
    __realloc_hook = GV.orealloc;
    __free_hook = GV.ofree;
    return;
}

static int load_sfile(const char *filename, Ulong base) {
    /* Loads all symbols from a file. You do not need -g to make a file have
    function name symbols; however you should not strip it! */
    char buf[MAXLNLEN];
    Ulong c = 0;
    FILE *fp;

    snprintf(buf, MAXLNLEN, "nm \"%s\" 2>/dev/null", filename);
    if((fp = popen(buf, "r")) == NULL) { return 0; }

    while(fgets(buf, MAXLNLEN, fp) != NULL) {
        char *addrp = buf, *type, *name;
        Ulong addr;

        if(*addrp == ' ') { continue; }
        if((type = strchr(buf, ' ')) == NULL) { continue; }
        *type++ = '\0';
        if(*type != 't' && *type != 'T') { continue; }
        // not interested in anything else than functions

        if((name = strchr(type, ' ')) == NULL) { continue; }
        *name++ = '\0';

        HX_chomp(name);
        addr = strtoul(addrp, NULL, 16);
        if(addr < base) { addr += base; }
        HXbtree_add(GV.stab, (void *)addr, name);
        ++c;
    }

    gprint(Flags.verbose, "Added %lu symbols from %s\n", c, filename);
    pclose(fp);
    return 1;
}

static const char *rsymbol(const void *addr) {
    /* Looks up the symbol for a specific address. load_sfile() hopefully
    filled the tree for us. */
    struct HXbtree *btree = GV.stab;
    struct HXbtree_node *node = btree->tree;
    const char *sym = NULL;
    const void *bp = NULL;
    int v;

    while(node != NULL) {
        v = btree->pcmp((void *)addr, node->key);
        if(node->key > bp && node->key <= addr) {
            bp  = node->key;
            sym = node->data;
        }
        if(v == 0) { break; }
        node = node->st[v > 0];
    }

    return sym;
}

static char *rsymbol_f(const void *addr) {
    const char *exec = NULL;
    const void *lowest;

    {
      struct HXbtrav *travp = HXbtrav_init(GV.amap, NULL);
      struct HXbtree_node *n = HXbtraverse(travp);
      if(n == NULL) { return NULL; }
      lowest = n->key;
      HXbtrav_free(travp);
    }

    {
      struct HXbtree *btree = GV.amap;
      struct HXbtree_node *node = btree->tree;
      const void *bp = 0x0;
      int v;

      while(node != NULL) {
          v = btree->pcmp((void *)addr, node->key);
          if(node->key > bp && node->key <= addr) {
              bp   = node->key;
              exec = node->data;
          }
          if(v == 0) { break; }
          node = node->st[v > 0];
      }

      if(bp != lowest) { addr -= (Ulong)bp; }
    }

    {
      char buf[MAXLNLEN], *p;
      FILE *fp;

      snprintf(buf, MAXLNLEN, "addr2line -e \"%s\" 0x%08lX",
       exec, (Ulong)addr);
      if((fp = popen(buf, "r")) == NULL) { return NULL; }

      p = fgets(buf, MAXLNLEN, fp);
      fclose(fp);
      if(p == NULL) { return NULL; }

      HX_chomp(buf);
      if(*buf == '?' || *buf == '\0') { return NULL; }
      if((p = strrchr(buf, '/')) != NULL) {
          return HX_strdup(++p);
      }

      return HX_strdup(buf);
    }
}

static void update_maps(void) {
    /* The only way I found to get to know of new symbols when someone calls
    dlopen() is to read the Linux-specific /proc/self/maps file. It works. */
    char buf[MAXFNLEN];
    FILE *fp = fopen("/proc/self/maps", "r");

    if(fp == NULL) { return; }

    while(fgets(buf, MAXLNLEN, fp) != NULL) {
        char *filename = strchr(buf, '/');
        if(filename == NULL) { continue; }
        HX_chomp(buf);
        if(HXbtree_find(GV.fmap, filename) == NULL) {
            /* Only load it if we have not yet. A hole remains open: What if
            someone dlopen()s, then dlclose()s, modifies the lib, and opens it
            again? (XXX) */
            Ulong base = strtoul(buf, NULL, 16);
            load_sfile(filename, base);
            HXbtree_add(GV.fmap, filename);
            HXbtree_add(GV.amap, (void *)base, filename);
        }
    }

    fclose(fp);
    return;
}

static int value_cmp(void *pa, void *pb) {
    return (pa > pb) - (pa < pb);
}

//==[ End of file ]============================================================
