/*=============================================================================
Vitalnix - User Management Suite
sysprog/slaned.c - ACCDB Server Lane Daemon
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2004
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
  $Id: slaned.c,v 1.3 2004/03/11 14:53:13 hirogen2 Exp $
=============================================================================*/
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libHX.h>
#include <accdb.h>

// Struct definitions et al
typedef uint8_t byte;

struct sv_unix_sock {
    struct sv_unix_sock *Next;
    char *path;
    struct sockaddr_un i;
    int fd;
};

struct sv_tcp_rule {
    struct sv_tcp_rule *Next;
    uint32_t netaddr, netmask;
};

struct sv_tcp_sock {
    struct sv_tcp_sock *Next;
    struct sv_tcp_rule *rules;
    in_port_t port;
    struct sockaddr_in i;
    int fd;
};

#define REQSIZE 1024
struct cl_conn {
    int fd;
    struct sockaddr r;
    socklen_t sz;

    int is_in_set;

    char *in_bufp, in_buf[REQSIZE];
    size_t in_fill, in_rest;
    int fetch_next_ok;

    char *outp, outq[REQSIZE];
    size_t out_fill, out_rest;
};

// Prototypes
static void ML_run(void);
static void *ML_serve(void *);
static void CL_new(int);
static void CL_read_gen(struct cl_conn *, struct HXlld_node *);
static void CL_write_gen(struct cl_conn *, struct HXlld_node *);
static void CL_close(struct HXlld_node *);
static struct accdb_module *DB_open(char *);
static void DB_close(struct accdb_module *);
static void SIG_setup(void);
static void SIG_block(void);
static void SIG_handler_hup(int);
static void SIG_handler_int(int);
static void SIG_handler_kill(int);
static void UNIX_gen_sockets(char *);
static int UNIX_gen_single(struct sv_unix_sock *);
static void UNIX_destroy(struct sv_unix_sock *);
static void TCP_gen_sockets(char *);
static int TCP_gen_single(struct sv_tcp_sock *);
static struct sv_tcp_sock *TCP_find(struct sv_tcp_sock *, in_port_t);
static void TCP_destroy_rt(struct sv_tcp_rule *);
static void TCP_destroy_st(struct sv_tcp_sock *);

static int make_netaddr(char *, uint32_t *);
static int make_netmask(char *, uint32_t *);
static void read_config(int);

// Variables
static struct {
    char *backend, *unix_binds, *tcp_binds;
} opt;

static struct HXlld *client_list = NULL;
static struct sv_unix_sock *listen_unix = NULL;
static struct sv_tcp_sock *listen_tcp = NULL;
static struct sv_tcp_rule *tcp_rules = NULL;
static struct accdb_module *xdb = NULL;

//-----------------------------------------------------------------------------
int main(int argc, char **argv) {
    printf("slaned v1.0.1\n---------------\n");
    read_config(0);
    client_list = HXlld_init(0);
    xdb = DB_open(opt.backend);
    UNIX_gen_sockets(opt.unix_binds);
    TCP_gen_sockets(opt.tcp_binds);
    SIG_setup();
    ML_run();

    TCP_destroy_st(listen_tcp);
    TCP_destroy_rt(tcp_rules);
    UNIX_destroy(listen_unix);
    HXlld_free(client_list);
    DB_close(xdb);
    read_config(1);
    printf("Exiting...\n");
    return EXIT_SUCCESS;
}

//-----------------------------------------------------------------------------
// Request handling
static pthread_mutex_t client_list_lock = PTHREAD_MUTEX_INITIALIZER;
static int ML_running = 1, ML_sync = 0;

static void ML_run(void) {
    struct sv_tcp_sock *l_tcp;
    struct sv_unix_sock *l_unix;
    pthread_t TH_serve;

    pthread_create(&TH_serve, NULL, ML_serve, NULL);
    fprintf(stderr, "Entering main loop\n");

    while(ML_running) {
        int highest = -1, rv;
        fd_set watch;
        FD_ZERO(&watch);

        l_tcp = listen_tcp;
        l_unix = listen_unix;
        while(l_tcp != NULL) {
            FD_SET(l_tcp->fd, &watch);
            if(l_tcp->fd > highest) { highest = l_tcp->fd; }
            l_tcp = l_tcp->Next;
        }
        while(l_unix != NULL) {
            FD_SET(l_unix->fd, &watch);
            if(l_unix->fd > highest) { highest = l_unix->fd; }
            l_unix = l_unix->Next;
        }

        rv = select(highest + 1, &watch, NULL, NULL, NULL);
        if(rv == -1 && errno == EINTR) { continue; }
        if(rv < 0) { ML_running = 0; break; }

        l_tcp = listen_tcp;
        l_unix = listen_unix;
        while(l_tcp != NULL) {
            if(FD_ISSET(l_tcp->fd, &watch)) { CL_new(l_tcp->fd); }
            l_tcp = l_tcp->Next;
        }
        while(l_unix != NULL) {
            if(FD_ISSET(l_unix->fd, &watch)) { CL_new(l_unix->fd); }
            l_unix = l_unix->Next;
        }
    }

    fprintf(stderr, "Exiting main loop\n");
    pthread_cancel(TH_serve);
    pthread_join(TH_serve, NULL);
    return;
}

static void *ML_serve(void *arg) {
    SIG_block();

    while(ML_running) {
        struct HXlld_node *travp;
        fd_set watch_rd, watch_wd;
        int highest = -1, rv;

        while(client_list->itemcount == 0) { usleep(100000); continue; }

        FD_ZERO(&watch_rd);
        FD_ZERO(&watch_wd);

        travp = client_list->first;
        while(travp != NULL) {
            struct cl_conn *cur = travp->ptr;
            if(cur->in_rest > 0) { FD_SET(cur->fd, &watch_rd); }
            if(cur->out_fill > 0) { FD_SET(cur->fd, &watch_wd); }
            if(cur->fd > highest) { highest = cur->fd; }
            cur->is_in_set = 1;
            travp = travp->Next;
        }

        rv = select(highest + 1, &watch_rd, &watch_wd, NULL, NULL);
        if(rv == -1 && errno == EINTR) { continue; }
        if(rv < 0) { perror("select()"); ML_running = 0; break; }

        travp = client_list->first;
        while(travp != NULL) {
            struct cl_conn *cur = travp->ptr;
            if(cur->is_in_set) {
                if(FD_ISSET(cur->fd, &watch_rd)) { CL_read_gen(cur, travp); }
                if(FD_ISSET(cur->fd, &watch_wd)) { CL_write_gen(cur, travp); }

                if(cur->fetch_next_ok) {
                }
            }
            cur->is_in_set = 0;
            travp = travp->Next;
        }
    }

    return NULL;
}

//-----------------------------------------------------------------------------
// Client handling

static void CL_new(int svfd) {
    struct cl_conn *new = NULL;
    int fd = accept(svfd, &new->r, &new->sz);
    if(fd < 0 || (new = malloc(sizeof(struct cl_conn))) == NULL) {
        return;
    }

    new->fd = fd;
    new->is_in_set = 0;

    new->in_bufp = new->in_buf;
    new->in_fill = 0;
    new->in_rest = REQSIZE;
    new->fetch_next_ok = 1;

    new->out_rest = REQSIZE;
//    cl_send_str(new, "SLANED p=1\n");

    pthread_mutex_lock(&client_list_lock);
    HXlld_push(client_list, new);
    pthread_mutex_unlock(&client_list_lock);
    return;
}

static void CL_read_gen(struct cl_conn *x, struct HXlld_node *t) {
    int s = recv(x->fd, x->in_bufp, x->in_rest, 0);
    if(s == 0) { CL_close(t); }
    x->in_bufp += s;
    x->in_fill += s;
    x->in_rest -= s;
    return;
}

static void CL_write_gen(struct cl_conn *x, struct HXlld_node *t) {
    return;
}

static void CL_close(struct HXlld_node *t) {
    struct cl_conn *conn = t->ptr;
    close(conn->fd);
    pthread_mutex_lock(&client_list_lock);
    HXlld_del(t);
    pthread_mutex_unlock(&client_list_lock);
    return;
}

//-----------------------------------------------------------------------------
// ACCDB stuff

static struct accdb_module *DB_open(char *module) {
    struct accdb_module *db;
    int eax;

    printf("Opening ACCDB (%s)\n", module);

    if(strcmp(module, "*") == 0 || strcmp(module, "slane") == 0) {
        fprintf(stderr,
          "\tWarning: You are using either \"*\" or the SLANE backend,\n"
          "\twhich might yield a recursive feedback loop.\n"
        );
    }

    if((db = accdb_load(module)) == NULL) {
        fprintf(stderr, "\tCould not load ACCDB back-end module: %s\n",
         strerror(errno));
        abort();
    }

    eax = db->sysctl(ACCDB_LOCK_EXCL);
    if(eax == -ENOSYS) {
        fprintf(stderr, "\tThe selected back-end doesnot support exclusive locking. Continuing at own risk.\n");
    } else if(eax <= 0) {
        fprintf(stderr, "\tCould not lock the back-end database.\n");
        accdb_unload(db);
        abort();
    }

    if((eax = db->open(&db->state)) <= 0) {
        fprintf(stderr, "\tCould not open ACCDB back-end: %s"
         " (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
        accdb_unload(db);
        abort();
    }

    db->sysctl(ACCDB_DELFLAGS, db->state, ACCDB_AFLUSH);
    return db;
}

static void DB_close(struct accdb_module *db) {
    db->sysctl(ACCDB_DELFLAGS, db->state, ACCDB_AFLUSH);
    db->sysctl(ACCDB_FLUSHDB);
    db->close(db->state);
    db->sysctl(ACCDB_UNLOCK_EXCL);
    return;
}

//-----------------------------------------------------------------------------
// Signal stuff

static void SIG_setup(void) {
    struct sigaction sa_hup, sa_int;
    sa_hup.sa_handler = SIG_handler_hup;
    sa_hup.sa_flags = SA_RESTART;
    sa_int.sa_handler = SIG_handler_int;
    sa_int.sa_flags = SA_RESTART;
    if(sigaction(SIGHUP, &sa_hup, NULL) < 0 ||
     sigaction(SIGINT, &sa_int, NULL) < 0 ||
     sigaction(SIGTERM, &sa_int, NULL) < 0) {
        fprintf(stderr, "SGN_setup(): %s\n", strerror(errno));
        abort();
    }
    printf("Signals set up\n");
    return;
}

static void SIG_handler_hup(int s) {
    printf("Received HUP signal, syncing database\n");
    ML_sync = 1;
    return;
}

static void SIG_block(void) {
    sigset_t s;
    sigfillset(&s);
    pthread_sigmask(SIG_SETMASK, &s, NULL);
    return;
}

static void SIG_handler_int(int s) {
    struct sigaction sa;
    sa.sa_handler = SIG_handler_kill;
    sa.sa_flags = SA_RESTART;
    printf("Received INT/TERM signal, shutting down...\n");
    ML_running = 0;
    sigaction(SIGINT, &sa, NULL);
    return;
}

static void SIG_handler_kill(int s) {
    printf("Received INT/TERM signal, killing daemon...\n");
    kill(0, SIGKILL);
    return;
}

//-----------------------------------------------------------------------------
// TCP socket handling code

static void TCP_gen_sockets(char *list) {
    struct sv_tcp_sock *st = (struct sv_tcp_sock *)&listen_tcp;
    char *host_s, *wp = list;

    printf("Setting up TCP sockets\n");
    while(st->Next != NULL) { st = st->Next; }

    while((host_s = HX_strsep(&wp, ",")) != NULL) {
        char *netmask_s = strrchr(host_s, '/'), *port_s = strrchr(host_s, ':');
        uint32_t netaddr_n, netmask_n = 0xFFFFFFFF;
        struct sv_tcp_sock *sp;
        struct sv_tcp_rule *rp;
        in_port_t port_n = 388;
        int is_new = 0;

        if(netmask_s != NULL) { *netmask_s++ = '\0'; }
        if(port_s != NULL) { *port_s++ = '\0'; }

        if((!make_netaddr(host_s, &netaddr_n)) ||
         (netmask_s != NULL && !make_netmask(netmask_s, &netmask_n)) ||
         (port_s != NULL && (port_n = strtoul(port_s, NULL, 0)) == 0)) {
            fprintf(stderr, "Invalid entry: %s\n", host_s);
            continue;
        }

        if((sp = TCP_find(listen_tcp, port_n)) == NULL) {
            sp = malloc(sizeof(struct sv_tcp_sock));
            sp->Next = NULL;
            sp->port = port_n;
            ++is_new;
            if(TCP_gen_single(sp) <= 0) {
                fprintf(stderr, "TCP_gen_single(): %s/%s:%s: %s\n",
                 host_s, netmask_s, port_s, strerror(errno));
                free(sp);
                continue;
            }
        }

        rp = (struct sv_tcp_rule *)&sp->rules;
        while(rp->Next != NULL) { rp = rp->Next; }
        rp->Next = malloc(sizeof(struct sv_tcp_rule));
        rp = rp->Next;
        rp->Next = NULL;
        rp->netaddr = netaddr_n & netmask_n;
        rp->netmask = netmask_n;

        if(is_new) { st = (st->Next = sp); }
        printf("\t- listening on TCP:%s/%s:%s\n",
         host_s, netmask_s, port_s);
    }

    return;
}

static int TCP_gen_single(struct sv_tcp_sock *d) {
    const int y = 1;
    d->i.sin_family = AF_INET;
    d->i.sin_port = htons(d->port);
    d->i.sin_addr.s_addr = INADDR_ANY;
    memset(&d->i.sin_zero, 0, sizeof(d->i.sin_zero));

    if((d->fd = socket(PF_INET, SOCK_STREAM, 0)) < 0 ||
     setsockopt(d->fd, SOL_SOCKET, SO_REUSEADDR, &y, sizeof(y)) < 0 ||
     bind(d->fd, (struct sockaddr *)&d->i, sizeof(struct sockaddr)) < 0 ||
     listen(d->fd, SOMAXCONN) < 0) {
        return -errno;
    }

    return 1;
}

static struct sv_tcp_sock *TCP_find(struct sv_tcp_sock *st, in_port_t port) {
    while(st != NULL) {
        if(st->port == port) { return st; }
        st = st->Next;
    }
    return NULL;
}

static void TCP_destroy_st(struct sv_tcp_sock *st) {
    struct sv_tcp_sock *Prev = NULL;
    printf("Removing TCP sockets\n");
    while(st != NULL) {
        if(Prev != NULL) { free(Prev); }
        close(st->fd);
        Prev = st;
        st = st->Next;
    }
    if(Prev != NULL) { free(Prev); }
    return;
}

static void TCP_destroy_rt(struct sv_tcp_rule *rt) {
    struct sv_tcp_rule *Prev = NULL;
    while(rt != NULL) {
        if(Prev != NULL) { free(Prev); }
        Prev = rt;
        rt = rt->Next;
    }
    if(Prev != NULL) { free(Prev); }
    return;
}

//-----------------------------------------------------------------------------
// UNIX socket handling code

static void UNIX_gen_sockets(char *list) {
    struct sv_unix_sock *travp = (struct sv_unix_sock *)&listen_unix;
    char *path, *wp = list;

    printf("Setting up UNIX sockets\n");
    while(travp->Next != NULL) { travp = travp->Next; }

    while((path = HX_strsep(&wp, ":")) != NULL) {
        struct sv_unix_sock *sp;
        if((sp = malloc(sizeof(struct sv_unix_sock))) == NULL) {
            perror("malloc()");
            abort();
        }

        sp->path = HX_strdup(path);
        sp->Next = NULL;

        if(UNIX_gen_single(sp) <= 0) {
            perror("UNIX_gen_single()");
            free(sp);
            continue;
        }

        travp->Next = sp;
        travp = travp->Next;
        fprintf(stderr, "\t- listening on UNIX:%s\n", sp->path);
    }

    return;
}

static int UNIX_gen_single(struct sv_unix_sock *d) {
    d->i.sun_family = AF_UNIX;
    strncpy(d->i.sun_path, d->path, sizeof(d->i.sun_path));
    unlink(d->path);

    if((d->fd = socket(PF_UNIX, SOCK_STREAM, 0)) < 0 ||
     bind(d->fd, (struct sockaddr *)&d->i, sizeof(struct sockaddr_un)) < 0 ||
     listen(d->fd, SOMAXCONN) < 0) {
         return -errno;
    }

    return 1;
}

static void UNIX_destroy(struct sv_unix_sock *st) {
    struct sv_unix_sock *Prev = NULL;
    printf("Removing UNIX sockets\n");
    while(st != NULL) {
        if(Prev != NULL) { free(Prev); }
        close(st->fd);
        free(st->path);
        Prev = st;
        st = st->Next;
    }
    if(Prev != NULL) { free(Prev); }
    return;
}

//-----------------------------------------------------------------------------
static int make_netaddr(char *addr, uint32_t *rp) {
    unsigned int ad[4];
    if(strchr(addr, '.') == NULL) { return 0; }

    sscanf(addr, "%u.%u.%u.%u", &ad[0], &ad[1], &ad[2], &ad[3]);

    *rp = ((ad[0] & 0xFF) << 24) | ((ad[1] & 0xFF) << 16) |
     ((ad[2] & 0xFF) << 8) | (ad[3] & 0xFF);
    return 1;
}

static int make_netmask(char *addr, uint32_t *rp) {
    if(strchr(addr, '.') != NULL) { return make_netaddr(addr, rp); }
    *rp = ~((1 << (32 - strtoul(addr, NULL, 0))) - 1);
    return 1;
}

static void read_config(int f) {
    const char *paths[] = {"sysprog/../vetc", "/etc", NULL};
    struct rconfig_opt options_table[] = {
        {"BACKEND",   RCONF_STRING, &opt.backend,    NULL, NULL},
        {"TCP_BIND",  RCONF_STRING, &opt.tcp_binds,  NULL, NULL},
        {"UNIX_BIND", RCONF_STRING, &opt.unix_binds, NULL, NULL},
        {NULL},
    };
    if(f) { accdb_rconfig_free(options_table); }
    else { accdb_rconfig_pv(paths, "slaned.cfg", options_table, RCONFIG_ONE); }
    return;
}

//==[ End of file ]============================================================
